package ru.dbotthepony.mc.otm.core.util

import java.util.concurrent.atomic.AtomicInteger
import java.util.concurrent.locks.ReentrantLock

class AtomicallyInvalidatedLazy<V>(private val invalidator: AtomicInteger, private val initializer: () -> V) : Lazy<V> {
	@Volatile
	private var thisCounter = -1
	@Volatile
	private var stored: Any? = Companion
	private val lock = ReentrantLock()

	override val value: V get() {
		if (thisCounter != invalidator.get()) {
			lock.lock()
			this.stored = Companion
			thisCounter = invalidator.get()
			lock.unlock()
		}

		var stored = stored

		if (stored !== Companion)
			return stored as V

		lock.lock()

		try {
			stored = initializer.invoke()
			this.stored = stored
			return stored
		} finally {
			lock.unlock()
		}
	}

	override fun isInitialized(): Boolean {
		return stored !== Companion && thisCounter == invalidator.get()
	}

	private companion object
}
